//  RW Foundations
// ================
// The RW Foundational Libraries

// Filename: ./include/incls_all.hpp
// -The following lines pull in the project's real *.hpp files.
// -note: this file also acts as a location for a few utility features

#pragma once

#include <cstdio>
#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <string>

using std::cout;
using std::exception;
using std::make_unique;
using std::ostringstream;
using std::printf;
using std::puts;
using std::runtime_error;
using std::string;
using std::unique_ptr;

//
// NOTE: It is probably a good idea to be cautious about moving things around
// hereafter within this file; some dependencies here are sensitive to ordering.
//

//------------------------------------------------------------------------------
/** Our alias for a @c void*
 */
using Address = void*;

//------------------------------------------------------------------------------
/** Treat an @c Address as a reference to a general type @a T
 * @param[in] pt the type's @c Address to take the reference to
 * @return @a pt 's reference
 */
template <class T>
T& reference_to(Address pt) {
  return *static_cast<T*>(pt);
}

//------------------------------------------------------------------------------
// THE RW 'SWITCHBOARD':
//

// clang-format off
namespace rw            { class RW; }
// clang-format on

using rw::RW;

///////
// TEST:
//
// clang-format off
namespace rw::pingpong  { class Pang; }
namespace rw::pingpong  { class Pang_b; }

namespace rw::pingpong  { class Peng; }
namespace rw::pingpong  { class Peng_b; }

namespace rw::pingpong  { class Ping; }
namespace rw::pingpong  { class Ping_b; }

namespace rw::pingpong  { class Pong; }
namespace rw::pingpong  { class Pong_b; }
// clang-format on

using rw::pingpong::Pang;
using rw::pingpong::Pang_b;

using rw::pingpong::Peng;
using rw::pingpong::Peng_b;

using rw::pingpong::Ping;
using rw::pingpong::Ping_b;

using rw::pingpong::Pong;
using rw::pingpong::Pong_b;

//------------------------------------------------------------------------------
/** Provides an identification structure for instantiated RW objects
 * @note: @b T ag, @b I d, @b P ointer
 *
 * This is simply a tagging-mechanism intended to be suitable for understanding
 * the potentially highly-complex callgraphs creatable inside deep RW (and
 * other) systems. IE, @a 'who-called-whom-and-who-answered-the-call?'.
 */
struct Tip {
  std::string   tag{"nobody"};  // sentinel for an unassigned
  unsigned long id{9001'42};    //  "
  RW*           ptr{nullptr};   //  "
};

//------------------------------------------------------------------------------
/** Prints a Tip item's elements out to the console, with an optional message
 * @param[in] tip the Tip data struct to print
 * @param[in] msg optional description (&tc) string
 */
void prt_tag(Tip const& tip, std::string const& msg = "");

//------------------------------------------------------------------------------
#include "rw/rw_incls_all.hpp"  // spcfc. #include'd before the other rw classes

///////
// TEST:
//
#include "pingpong/pingpong_incls_all.hpp"

//------------------------------------------------------------------------------

inline void error(std::string const& s) {
  ostringstream ss{};
  ss << s << '\n';
  throw runtime_error(ss.str());
}

inline void error(std::string const& s, int const i) {
  ostringstream ss{};
  ss << s << ": " << i << '\n';
  throw runtime_error(ss.str());
}

inline void error(std::string const& s1, std::string const& s2) {
  ostringstream ss{};
  ss << s1 << ": " << s2 << '\n';
  throw runtime_error(ss.str());
}

//    --=^=--

// "Is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v211121
// Copyright (2021)
// License MIT(Expat) https://opensource.org/licenses/MIT
